#include <stdio.h>
#include "main.h"
#include <string.h>

typedef unsigned char u8;

#define MIN_RUN     3
#define MAX_RUN     (127 + MIN_RUN)
#define MIN_STRING  1
#define MAX_STRING  (127 + MIN_STRING)

void compress(u8* input, int& inputSize, u8* output, int& outputSize)
{
    int stringStart = 0;
    int runStart = 0;
    int currentByte = 1;
    outputSize = 0;
    bool isInRun = false;

    while (true)
    {
        if (isInRun == false)
        {
            // string mode
            if (input[currentByte] != input[runStart])
                runStart = currentByte;

            int stringLength = runStart - stringStart;
            int runLength    = currentByte + 1 - runStart;

            if (currentByte == inputSize ||
                runLength    == MIN_RUN ||
                stringLength == MAX_STRING)
            {
                // emit string
                if (stringLength >= MIN_STRING)
                {
                    output[outputSize++] = stringLength - MIN_STRING;

                    for (int i = 0; i < stringLength; i++)
                        output[outputSize++] = input[stringStart++];
                }

                // go to run mode
                if (runLength == MIN_RUN)
                    isInRun = true;
            }
        }
        else
        {
            // run mode
            stringStart = currentByte;

            int runLength = currentByte - runStart;

            if (input[currentByte] != input[runStart] ||
                currentByte == inputSize ||
                runLength == MAX_RUN)
            {
                // emit run
                output[outputSize++] = 128 + runLength - MIN_RUN;
                output[outputSize++] = input[runStart];
                runStart = currentByte;

                // go to string mode
                isInRun = false;
            }
        }

        if (currentByte == inputSize)
            break;

        currentByte++;
    }
}

void uncompress(u8* input, int& inputSize, u8* output, int& outputSize)
{
    int i = 0;
    outputSize = 0;

    while (i != inputSize)
    {
        if (input[i] < 128)
        {
            int length = input[i++] + MIN_STRING;

            for (int j = 0; j < length; j++)
                output[outputSize++] = input[i++];
        }
        else
        {
            int length = input[i++] - 128 + MIN_RUN;

            for (int j = 0; j < length; j++)
                output[outputSize++] = input[i];

            i++;
        }
    }
}

int main(int argc, char* argv[])
{
    u8  input[] = "ABCDAAAABBCDDDDEEEEE";
    int inputSize = strlen((char*)input);
    u8  output[inputSize * 2];
    int outputSize = 0;

    printf("input:  %s\n", input);
    printf("input size: %d\n", inputSize);

    compress(input, inputSize, output, outputSize);

    printf("output: ");

    int i = 0;
    while (i != outputSize)
    {
        printf("(%d)", output[i]);

        if (output[i] < 128)
        {
            int length = output[i++] + MIN_STRING;

            for (int j = 0; j < length; j++)
                printf("%c", output[i++]);
        }
        else
        {
            i++;
            printf("%c", output[i++]);
        }

        printf(" ");
    }

    printf("\n");
    printf("output size: %d\n", outputSize);

    uncompress(output, outputSize, input, inputSize);

    printf("uncomp: %s\n", input);
    printf("uncompSize: %d\n", inputSize);

    return 0;
}
